# -*- perl -*-
#
# Copyright (C) 1996 Ken MacLeod
# See the file COPYING for distribution terms.
#

use strict;

use Text::GDO::Node;
use Text::GDO::Reference;

# XXX formatters not used by other formatters should be loaded
# dynamically
use Text::GDO::toDump;
use Text::GDO::toAscii;
use Text::GDO::toHTML;

package Text::GDO::DocObject;
@Text::GDO::DocObject::ISA = qw{Text::GDO::Node};

sub to {
    my ($self, $output_type, %args) = @_;
    my ($arg, @args);

    if (defined $args{'args'}) {
	foreach $arg (split (/;/, $args{'args'})) {
	    my ($argname, $value) = split (/=/, $arg, 2);
	    if (!defined $value) {
		if ($argname =~ /^no/i) {
		    $value = 0;
		    $argname = substr ($argname, 2);
		} else {
		    $value = 1;
		}
	    }
	    push (@args, $argname, $value);
	}
    } else {
	@args = ();
    }

    my ($output_instance) = eval "\$self->To$output_type(\@args)"
	|| die "$::prog: could not output document, $@\n";

    return ($output_instance);
}

sub content {
    my ($self) = @_;

    if (defined $self->{'content'}) {
	return @{$self->{'content'}};
    } else {
	return ();
    }
}

sub as_string {
    my ($self, $context) = @_;
    my ($data) = "";
    my ($block);

    foreach $block (@{$self->{'content'}}) {
	$data .= $block->as_string ($context);
    }

    return $data;
}

sub is_inline {
    return 0;
}

package Text::GDO::Document;
@Text::GDO::Document::ISA = qw{Text::GDO::DocObject};

#
# `Author' is based loosely on the `vCard' specification,
# http://www.versit.com/
#
package Text::GDO::Author;
@Text::GDO::Author::ISA = qw{Text::GDO::DocObject};

sub name {
    my ($self) = @_;

    # XXX not only are these very incomplete, but they should be
    # returning objects, not strings
    # XXX need a way to return a "space" Data object
    if ($self->{'formattedname'}) {
	return $self->{'formattedname'}->as_string (Text::GDO::Context::string_context());
    } elsif ($self->{'givenname'} || $self->{'familyname'}) {
	my (@names);
	push (@names, $self->{'givenname'}->as_string (Text::GDO::Context::string_context()))
	    if ($self->{'givenname'});
	push (@names, $self->{'familyname'}->as_string (Text::GDO::Context::string_context()))
	    if ($self->{'familyname'});
	return join (" ", @names);
    } elsif ($self->{'othername'}) {
	return $self->{'othername'}->as_string (Text::GDO::Context::string_context());
    } elsif ($self->{'content'}) {
	return $self->as_string (Text::GDO::Context::string_context());
    }

    return undef;
}

sub address {
    my ($self) = @_;
    my ($str) = undef;

    $str .= $self->{'postofficeaddress'}->as_string (Text::GDO::Context::string_context()) . "\n"
	if (defined $self->{'postofficeaddress'});

    $str .= $self->{'street'}->as_string (Text::GDO::Context::string_context()) . "\n"
	if (defined $self->{'street'});

    my ($locality, $region, $postalcode);
    $locality = $self->{'locality'}->as_string (Text::GDO::Context::string_context())
	if (defined $self->{'locality'});
    $region = $self->{'region'}->as_string (Text::GDO::Context::string_context())
	if (defined $self->{'region'});
    $postalcode = $self->{'postalcode'}->as_string (Text::GDO::Context::string_context())
	if (defined $self->{'postalcode'});

    $str .= $locality   if (defined $locality);
    $str .= ", "        if (defined $locality && defined $region);
    $str .= $region     if (defined $region);
    $str .= "  "        if (defined $locality || defined $region);
    $str .= $postalcode if (defined $postalcode);
    $str .= "\n"        if (defined $locality || defined $region
			    || defined $postalcode);

    $str .= $self->{'country'}->as_string (Text::GDO::Context::string_context()) . "\n"
	if (defined $self->{'country'});

    return $str;
}
    
package Text::GDO::Inline;
@Text::GDO::Inline::ISA = qw{Text::GDO::DocObject};

sub is_inline {
    return 1;
}

package Text::GDO::Inline_Quote;       @Text::GDO::Inline_Quote::ISA = qw{Text::GDO::Inline};
package Text::GDO::Inline_Emphasis;    @Text::GDO::Inline_Emphasis::ISA = qw{Text::GDO::Inline};
package Text::GDO::Inline_Literal;     @Text::GDO::Inline_Literal::ISA = qw{Text::GDO::Inline};
package Text::GDO::Inline_Replaceable; @Text::GDO::Inline_Replaceable::ISA = qw{Text::GDO::Inline};
package Text::GDO::Inline_Package;     @Text::GDO::Inline_Package::ISA = qw{Text::GDO::Inline};
package Text::GDO::Inline_Index;       @Text::GDO::Inline_Index::ISA = qw{Text::GDO::Inline};

package Text::GDO::List;
@Text::GDO::List::ISA = qw{Text::GDO::DocObject};

sub first_item_num {
    my ($self) = @_;
    my ($number) = 1;

    $number = $self->prev_with_gi('List')->last_item_num
	if (defined $self->{'continued'});

    return $number;
}

sub last_item_num {
    my ($self) = @_;
    my ($number) = $self->first_item_num;

    return $number + $#{$self->{'content'}} + 1;
}

sub type {
    my ($self) = @_;

    return $self->{'type'};
}
	    
package Text::GDO::ListItem;
@Text::GDO::ListItem::ISA = qw{Text::GDO::DocObject};

sub number {
    my ($self) = @_;
    my ($number) = $self->{'parent'}->first_item_num;
    my ($node);

    foreach $node (@{$self->{'parent'}->{'content'}}) {
	last if $node == $self;

	# XXX $node->is_listitem
	$number ++
	    if (ref $node eq 'Text::GDO::ListItem');
    }

    return $number;
}

package Text::GDO::ListTerm;   @Text::GDO::ListTerm::ISA = qw{Text::GDO::DocObject};
package Text::GDO::Block;      @Text::GDO::Block::ISA = qw{Text::GDO::DocObject};
package Text::GDO::Paragraph;  @Text::GDO::Paragraph::ISA = qw{Text::GDO::Block};
package Text::GDO::BlockQuote; @Text::GDO::BlockQuote::ISA = qw{Text::GDO::Block};
package Text::GDO::Line;       @Text::GDO::Line::ISA = qw{Text::GDO::Block};
package Text::GDO::NoFill;     @Text::GDO::NoFill::ISA = qw{Text::GDO::Block};
package Text::GDO::Screen;     @Text::GDO::Screen::ISA = qw{Text::GDO::Block};

package Text::GDO::Section;
@Text::GDO::Section::ISA = qw{Text::GDO::DocObject};
# XXX Preface and Appendix broken-out to subclasses?

sub formal_name {
    my ($self) = @_;

    $self->type . " " . $self->formal_number;
}

sub formal_number {
    my ($self) = @_;

    # XXX make into a settable parameter
    my ($num_sects) = 9999;
    my (@sect_nums, $node);
    my (@rootpath) = $self->rootpath;
    # XXX the following is an example of why `rootpath' (among other
    # selectors) should take a query argument
    foreach $node (@rootpath) {
	# XXX $node->is ('section')
	if (ref $node eq 'Text::GDO::Section') {
	    my ($node_type) = $node->type;
	    # XXX $node->is ('preface')
	    if ($node_type eq 'Preface') {
		return "";
	    }
	    last;
	}
    }
    foreach $node (@rootpath) {
	if (ref $node eq 'Text::GDO::Section') {
	    push (@sect_nums, $node->number);
	    last if ($num_sects-- == 0)
	}
    }

    return join ('.', @sect_nums);
}

sub number {
    my ($self) = @_;
    my ($number) = 1;
    my ($in_appendices) = 0;
    my ($node);

    foreach $node (@{$self->{'parent'}->{'content'}}) {
	last if $node == $self;

	# XXX $node->is ('section') && !$node->is ('preface')
	if (ref $node eq 'Text::GDO::Section'
	    && $node->type ne 'Preface') {
	    $number ++;
	    if (!$in_appendices && $node->type eq 'Text::GDO::Appendix') {
		$in_appendices = 1;
		$number = "A";
	    }
	}
    }

    return $number;
}

sub level {
    my ($self) = @_;

    my ($sect_level) = 0;
    my ($node);
    foreach $node ($self->rootpath) {
	if (ref $node eq 'Text::GDO::Section') {
	    $sect_level ++;
	}
    }

    return $sect_level;
}

sub type {
    my ($self) = @_;

    if (defined $self->{'type'}) {
	return $self->{'type'};
    } else {
	return "Section";
    }
}
	    
package Text::GDO::Formal;
@Text::GDO::Formal::ISA = qw{Text::GDO::DocObject};

sub formal_name {
    my ($self) = @_;

    return $self->{'type'} . " " . $self->formal_number;
}

sub formal_number {
    my ($self) = @_;

    # XXX make into a settable parameter
    my ($num_sects) = 1;
    my (@sect_nums, $node);
    foreach $node ($self->rootpath) {
	# XXX $node->is ('section')
	if (ref $node eq 'Text::GDO::Section') {
	    push (@sect_nums, $node->number);
	    last if ($num_sects-- == 0)
	}
    }

    return join ('.', @sect_nums) . "-" . $self->number;
}

sub number {
    my ($self) = @_;
    my ($number) = 1;
    my ($node);

    foreach $node (@{$self->{'parent'}->{'content'}}) {
	last if $node == $self;

	# XXX $node->is ('formal') && $node->is ($self->type)
	if (ref $node eq 'Text::GDO::Formal'
	    && $node->{'type'} eq $self->{'type'}) {
	    $number ++;
	}
    }

    return $number;
}

sub type {
    my ($self) = @_;

    return $self->{'type'};
}
	    
package Text::GDO::Bridge;
@Text::GDO::Bridge::ISA = qw{Text::GDO::DocObject};

package Text::GDO::Admonition;
@Text::GDO::Admonition::ISA = qw{Text::GDO::DocObject};

sub type {
    my ($self) = @_;

    return $self->{'type'};
}
	    
package Text::GDO::Table;
@Text::GDO::Table::ISA = qw{Text::GDO::DocObject};

sub num_parts {
    my ($self) = @_;

    my (@content) = $self->content;
    return $#content + 1;
}

sub num_rows {
    my ($self) = @_;
    my ($rows) = 0;
    my ($part);

    if (!defined $self->{'num_rows'}) {
	foreach $part ($self->parts) {
	    $rows += $part->num_rows;
	}
	$self->{'num_rows'} = $rows;
    }

    return $self->{'num_rows'};
}

sub num_columns {
    my ($self) = @_;

    if (!defined $self->{'num_columns'}) {
	my ($max_columns) = 0;
	my ($columns, $part);

	foreach $part ($self->parts) {
	    $columns = $part->num_columns;
	    $max_columns = $columns
		if ($columns >$max_columns);
	}
	$self->{'num_columns'} = $max_columns;
    }

    return $self->{'num_columns'};
}

sub parts {
    my ($self) = @_;

    return $self->content;
}

package Text::GDO::TablePart;
@Text::GDO::TablePart::ISA = qw{Text::GDO::DocObject};

sub num_parts {
    return (1);
}

sub parts {
    return ($_[0]);
}

sub num_rows {
    my ($self) = @_;

    my (@content) = $self->content;
    return $#content + 1;
}

sub num_columns {
    my ($self) = @_;

    if (!defined $self->{'num_columns'}) {
	my ($max_columns) = 0;
	my ($columns, $row);

	foreach $row ($self->rows) {
	    $columns = $row->num_columns;
	    $max_columns = $columns
		if ($columns >$max_columns);
	}
	$self->{'num_columns'} = $max_columns;
    }

    return $self->{'num_columns'};
}

sub rows {
    my ($self) = @_;

    return $self->content;
}

package Text::GDO::TableRow;
@Text::GDO::TableRow::ISA = qw{Text::GDO::DocObject};

sub num_columns {
    my ($self) = @_;

    my (@content) = $self->content;
    return $#content + 1;
}

sub entries {
    my ($self) = @_;

    return $self->content;
}

package Text::GDO::TableEntry;  @Text::GDO::TableEntry::ISA = qw{Text::GDO::DocObject};
package Text::GDO::XRef;        @Text::GDO::XRef::ISA = qw{Text::GDO::Inline};
package Text::GDO::URL;         @Text::GDO::URL::ISA = qw{Text::GDO::XRef};
package Text::GDO::XRefEnd;     @Text::GDO::XRefEnd::ISA = qw{Text::GDO::XRef};

package Text::GDO::Data;
@Text::GDO::Data::ISA = qw{Text::GDO::Inline};

sub new {
    my ($type, %value) = @_;

    my ($self) = \%value;

    bless $self, $type;
}

sub as_string {
    my ($self, $context) = @_;

    if (ref $self->{'source'} eq "") {
	return $self->{'source'};
    } else {
	$self->{'source'}->as_string ($context);
    }
}

package Text::GDO::SGML::Event;

#
# Almost every `start_element' has already been mapped to a subclass
# of `DocObject', the only things really left are actual data,
# including entities, etc.
#
sub as_string {
    my ($event, $context) = @_;
    my ($data) = "";

    my ($line_end) = $context->fill ? " " : "\n";

    if ($event->type eq 'start_element') {
	my ($subevent);
	foreach $subevent (@{$event->element}) {
	    $data .= $subevent->as_string ($context);
	}
    } elsif ($event->type eq 'cdata') {
	my ($str) = $event->data;
	if ($::no_SGMLSpm) {
	    $str =~ s/\\\|(.*?)\\\|/$context->charent($1)/ge;
	    $str =~ s/\\([0-7][0-7]?[0-7]?|.)/&esc($1, $line_end)/eg;
	}
	$data .= $str;
    } elsif ($event->type eq 're') {
	$data .= $line_end;
    } elsif ($event->type eq 'entity') {
	# XXX handle notations
	warn "$::prog: notations not handled, reading entity \`"
	    . $event->data->name . "' in verbatim\n";
	# XXX there's a whole bunch of stuff not going on here
	my ($file_name);
	($file_name = $event->data->filenames->[0]) =~ s/^<.*>//;
	# XXX use cache
	$data .= read_file ($file_name);
    } elsif ($event->type eq 'sdata') {
	$data .= $context->charent($event->data);
    } else {
	$data .= "\n" . $event->type . "\n";
    }

    $data;
}

sub read_file {
    my ($file_name) = @_;
    my ($str);

    # Read a file in using Perl's "slurp" trick
    @ARGV = ($file_name);
    undef $/;
    $str = <>;
    $/ = "\n";

    return $str;
}

#
# translate `nsgmls' output escape sequences into just perl data
#
sub esc {
    my($char, $line_end) = @_;

    if ($char eq '012' || $char eq '12') {
	"";			# ignore RS
    } elsif ($char =~ /^[0-7]/) {
	sprintf("%c", oct ($char));
    } elsif ($char eq 'n') {
	$line_end;
    } elsif ($char eq '|') {
	"";
    } elsif ($char eq "\\") {
	"\\";
    } else {
	$char;
    }
}

1;
