# -*- perl -*-
#
# Copyright (C) 1996 Ken MacLeod
# See the file COPYING for distribution terms.
#

use strict;

use Text::GDO::Node;

# forward declare this
package Text::GDO::Context::Specification;

package Text::GDO::Specification;
@Text::GDO::Specification::ISA = qw{Text::GDO::Node};

#
# `package_index' generates unique IDs for each new transform package
# that is merged from specifications, see `merge'
#
$Text::GDO::Specification::package_index = 0;

# XXX these are from a prior prototype, not yet integrated
#$Text::GDO::Specification::entities{'[ctx   ]'} = '$instance->';
#$Text::GDO::Specification::entities{'[data  ]'} = '$instance->{content}';
#$Text::GDO::Specification::entities{'[gi    ]'} = '$element->name';
#$Text::GDO::Specification::entities{'[qi    ]'} = "\01";
$Text::GDO::Specification::entities{'[lt    ]'} = '<';
$Text::GDO::Specification::entities{'[amp   ]'} = '&';
#$Text::GDO::Specification::entities{'[tfm   ]'} = '$instance->{transform}->';

sub new {
    my ($type, $instance, %args) = @_;
    my ($self);

    my ($spec_file) = $args{'file'};
    if (!defined $spec_file) {
	$spec_file = Text::GDO->find_file ($instance->doc_spec . ".spec", $instance->{'file_dir'});
    }
    die "$::prog: error creating new specification: no file specified\n"
	if (!defined $spec_file);

    # XXX use caching code
    if (defined $Text::GDO::Specification::specs{$spec_file}) {
	$self = $Text::GDO::Specification::specs{$spec_file};
    } else {
	# load specification SGML instance
	my ($spec_instance) = new Text::GDO::SGML ('file' => $spec_file)
	    || die "$::prog: can't load \`$spec_file', exiting\n";

	# get a `spec_transform'
	my ($spec_transform) = new Text::GDO::Spec::Transform;

	# map specification instance to a specification object using
	# the specification transform (say that fast :-)
	$self = $spec_transform->map_instance($spec_instance)
	    || die "$::prog: can't map \`$spec_file', exiting\n";

	$Text::GDO::Specification::specs{$spec_file} = $self;
    }

    # returns a `Spec::Instance'
    $self->{'file'} = $spec_file;
    return $self;
}

#
# `merge' is a class method that merges multiple `Specification'
# objects into a single transform.  The resulting transform is a new
# Perl class that isa subclass of `Transform'
#
sub merge {
    my ($class, @spec_instances) = @_;

    # XXX create some index out of the multiple `specInstances' and
    # retrieve the merged spec from cache if it already exists
    if (defined $Text::GDO::Specifications::transforms{$spec_instances[0]->{'file'}}) {
	return $Text::GDO::Specifications::transforms{$spec_instances[0]->{'file'}};
    }

    my ($self) = {};
    $Text::GDO::Specifications::transforms{$spec_instances[0]->{'file'}} = $self;

    $self->{'specification_id'} = "Spec"
	. $Text::GDO::Specification::package_index ++;

    bless $self, $self->{'specification_id'};

    # XXX not really merged yet
    my ($spec_instance) = $spec_instances[0];

    my ($context) = new Text::GDO::Context::Specification (
				'entities' => \%Text::GDO::Specification::entities,
				'current' => [{}]);

    $context->push ('fill' => 0, 'rule_names' => []);
    my ($package) = $self->{'specification_id'};
    my ($perl_code) = <<"EOT";
# -*- perl -*-

use strict;

use Text::GDO::Transform;
use Text::GDO::DocObject;

package $package;
\@${package}::ISA = qw{Text::GDO::Transform};

sub new {
    my (\$type) = \@_;
    my (\$self) = {};
    bless \$self, '$package';
    \$self->init;
    return \$self;
}

sub init {
    my (\$self) = \@_;

    \$self->{'map'} = {
EOT

    my ($rule);
    foreach $rule (@{$spec_instance->{'rules'}}) {
	$perl_code .= $rule->mappings ($context);
    }

    $perl_code .= <<EOT;
    };

}
EOT
    $perl_code .= $spec_instance->{'stuff'}->as_string ($context)
        if (defined $spec_instance->{'stuff'});

    my ($tmp_file) = "/tmp/sroff-$$-$package";
    open (PERLCODE, ">$tmp_file")
        || die "$::prog: can't open \`$tmp_file' for writing: $!\n";
    print PERLCODE $perl_code;
    # add another `package' in case `stuff' changed it
    print PERLCODE "package $package;\n";
    print PERLCODE $spec_instance->to_perl ($context);
    print PERLCODE "1;\n";
    close (PERLCODE);

    if (!eval "require '$tmp_file'") {
	die "$::prog: Loading specification \`$tmp_file': $@\n";
    }
    # XXX need debug switch to keep
    unlink ($tmp_file);

    $context->pop;

    $self->init;

    return $self;
}

#
# `Spec::Object' is just a abstract class to subclass from `Node'
#
package Text::GDO::Spec::Object;
@Text::GDO::Spec::Object::ISA = qw{Text::GDO::Node};

package Text::GDO::Spec::Instance;
@Text::GDO::Spec::Instance::ISA = qw{Text::GDO::Spec::Object};

sub to_perl {
    my ($self, $context) = @_;
    my ($rule, $perl_code);

    foreach $rule (@{$self->{'rules'}}) {
	$perl_code .= $rule->to_perl ($context);
    }

    return $perl_code;
}

package Text::GDO::Spec::Rule;
@Text::GDO::Spec::Rule::ISA = qw{Text::GDO::Spec::Object};

#
# return the mapping that was generated from the generic identifier of
# this and all parent rules
#
sub mapped_gi {
    my ($self, $context) = @_;

    if (!defined $self->{'mapped_gi'}) {
	my ($query) = $self->{'query'}->as_string ($context);

	# use only the first query element
	my ($sub_name) = (split (/\s+/, $query))[0];

	# replace anything that's not `a-zA-Z0-9_' with `_'
	$sub_name =~ tr/a-zA-Z0-9_/_/cs;

	# the name of this function is a concatenation of all parent's names
	my ($function_name) = join ('_', @{$context->rule_names}, $sub_name);
	$self->{'mapped_gi'} = $function_name;
    }

    return ($self->{'mapped_gi'});
}

#
# return a string that contains perl code to map each generic
# identifier to it's mapped Perl name
#
sub mappings {
    my ($self, $context) = @_;
    my ($sub_name);
    my ($mappings) = "";

    my ($query) = $self->{'query'}->as_string ($context);
    my ($func_name) = $self->mapped_gi ($context);
    foreach $sub_name (split (/\s+/, $query)) {
	$mappings .= "        '$sub_name' => '$func_name',\n";
    }

    return $mappings;
}

#
# Convert a rule into Perl code
#
sub to_perl {
    my ($self, $context) = @_;
    my ($perl_code);

    my ($function_name) = $self->mapped_gi ($context);

    my ($perl_code) = <<EOT;
sub map$function_name {
    my (\$self, \$data, \$event, \$element) = \@_;
    my (\$parent) = \$data->{'diversion'}[-1];

EOT

    if (defined $self->{'rules'}) {
	my ($rule);
	$perl_code .= <<EOT;
    \$self->map_gi (
EOT
	$context->push_rule_names ($function_name);
        foreach $rule (@{$self->{'rules'}}) {
	    $perl_code .= $rule->mappings ($context);
	}
	$context->pop_rule_names;
        $perl_code .= <<EOT;
    );
EOT
    }

    my ($gen_object);
    if (defined $self->{'isa'}) {
	$gen_object = $self->{'isa'}->as_string ($context);
    } else {
	$gen_object = "Block";
    }
    if (defined $self->{'isa'} && $gen_object eq 'Holder') {
	$perl_code .= <<EOT;
    \$self->_map_instance (\$data, \$event);
EOT
    } else {	
	$perl_code .= <<EOT;
    my (\$gen_object) = new Text::GDO::$gen_object;
    \$gen_object->{'parent'} = \$parent;
    \$gen_object->{'source'} = \$event;
    push (\@{\$data->{'diversion'}}, \$gen_object);
    \$self->save_attr (\$data, \$event);
    \$self->_map_instance (\$data, \$event);
EOT

        if (defined $self->{'save'}) {
	    my ($save) = $self->{'save'}->as_string ($context);
	    $perl_code .= "    \$parent->$save = pop (\@{\$data->{'diversion'}});\n"
        } elsif (defined $self->{'push'}) {
	    my ($push) = $self->{'push'}->as_string ($context);
	    $perl_code .= "    push (\@{\$parent->$push}, pop (\@{\$data->{'diversion'}}));\n";
	} else {
	    $perl_code .= "    push (\@{\$parent->{'content'}}, pop (\@{\$data->{'diversion'}}));\n";
	}
    }

    if (defined $self->{'rules'}) {
	$perl_code .= <<EOT;
    \$self->unmap_gi;
EOT
    }

    $perl_code .= <<EOT;
}

EOT

    my ($rule);
    foreach $rule (@{$self->{'rules'}}) {
	$perl_code .= $rule->to_perl ($context);
    }

    return $perl_code;
}

#
# A `Thing' is just a generic holder of an element
#
package Text::GDO::Spec::Thing;
@Text::GDO::Spec::Thing::ISA = qw{Text::GDO::Spec::Object};

sub as_string {
    my ($self, $context) = @_;
    my ($data) = "";
    my ($block);

    foreach $block (@{$self->{'content'}}) {
	$data .= $block->as_string ($context);
    }

    return $data;
}

#
# An `Attr' is a reference to an element attribute, the perl code that
# is returned retrieves the attribute from the element of the current
# event.
#
package Text::GDO::Spec::Attr;
@Text::GDO::Spec::Attr::ISA = qw{Text::GDO::Spec::Object};

sub as_string {
    my ($self, $context) = @_;
    my ($data) = "";
    my ($block);

    foreach $block (@{$self->{'content'}}) {
	$data .= $block->as_string ($context);
    }

    # XXX no error checking, very likely to fail
    return "\$event->data->attribute('$data')->value";
}

#
# Spec's `Context' stores the hierarchy of rules, see `mappings'
#
package Text::GDO::Context::Specification;
@Text::GDO::Context::Specification::ISA = qw{Text::GDO::Context};

sub rule_names     { return $_[0]->{'current'}[-1]{'rule_names'} }

sub set_rule_names { return $_[0]->{'current'}[-1]{'rule_names'} = $_[1]; }

sub push_rule_names { return push (@{$_[0]->{'current'}[-1]{'rule_names'}}, $_[1]) }
sub pop_rule_names  { return pop (@{$_[0]->{'current'}[-1]{'rule_names'}}) }

1;
